 /*
 * File: rlp.cc
 * Date: March 28, 2004
 * Author: Vikas Paliwal
 * Email: vpaliwal@sce.carleton.ca
 * Implementation of an RLP agent
 */

#include <stdio.h>
#include <math.h>
#include <iostream.h>
#include <fstream>
#include "rlp.h"
#include "rtp.h"
#include "tcp.h"
#include "errmodel.h"

int hdr_rlp::off_rlp_;
int hdr_rlp::off_rtp_;
int hdr_rlp::off_tcp_;
NakListEntry* NakListEntry::free_;
//fstream RlpAgent::out_("delay");

static class RlpHeaderClass : public PacketHeaderClass
{
	public:
		  RlpHeaderClass() : PacketHeaderClass("PacketHeader/Rlp",sizeof(hdr_rlp))
		  {
				bind_offset(&hdr_rlp::off_rlp_);
				bind_offset(&hdr_rlp::off_rtp_);
				bind_offset(&hdr_rlp::off_tcp_);
		  }
	private:

} class_rlphdr;

static class RlpClass : public TclClass
{
	public:
		  RlpClass() : TclClass("Agent/Rlp") {}
		  TclObject* create(int, const char*const*)
		  {
	      return (new RlpAgent());
		  }
} class_rlp;


RlpAgent::RlpAgent() : Agent(PT_RLP), rlp_timer_(this)
{
   bind("rawrate_",  &rlprate_);
   bind("packetSize_", &size_);
   bind("rlp_delay", &RLP_DELAY);
   NUM_ROUNDS = 3;
   rlp_timer_.resched(0.02);
   nak_list_.init(this); 
   readyToSend = false;
   ifIdle = false;
   lvr = lvn = 1;
   lvs = 1;
   ipq = 0;
   ipPacket = 0;
   currIpPacket = 0;
   currSendIpPacket = 0;
   lastSeqNo = 0;
   last = 0;
   ipCounter = 0;
   ipCounter1 = 0;
   em = 0;
}

int RlpAgent::command(int argc, const char*const* argv)
 {

	if ( (argc == 3)&& (strcmp(argv[1], "dst") == 0))
	{
	     	Agent* a = (Agent*)TclObject::lookup(argv[2]);
       		dst=a;
       		return (TCL_OK);
     	}
	if ( (argc == 3)&& (strcmp(argv[1], "errmodel") == 0))
	{
		ErrorModel* e = (ErrorModel*)TclObject::lookup(argv[2]);
		em = e;
		return (TCL_OK);
	}
   return (Agent::command(argc, argv));
 }


 void RlpAgent::recv(Packet* pkt, Handler*)
 {
   hdr_ip* hdrip = hdr_ip::access(pkt);
   hdr_cmn* hdrcmn = hdr_cmn::access(pkt);
   packet_t typeOfPacket = hdrcmn->ptype();
   if( ifIdle &&em && ((typeOfPacket == PT_RLP) ||  (typeOfPacket == PT_RLP_NAK) || (typeOfPacket == PT_RLP_IDLE)))
   {
   	em->recv(pkt,(Handler*)0);
   }
  if(!ifIdle) ifIdle = true;
   if(typeOfPacket == PT_RLP_IDLE)
   {
		if(  ipPacket ) 
		{
			hdr_rlp* hdar = hdr_rlp::access(ipPacket);
			if(hdar->txmtd){
			lastSeqNo = hdar->l_seq;
			if(!hdar->sent && !hdar->corrupted)reassembleBuffer.enque(ipPacket); 
			readyToSend = true;
			ipPacket = 0;
			}
		}
		if(compare(lvr, (lastSeqNo+1)%SEQ_NO_SIZE)) 
		{
			for(int cnt = lvr; compare(cnt, (lastSeqNo+1)%SEQ_NO_SIZE);cnt  =  ((cnt+1)%SEQ_NO_SIZE))
			{	
				if(!nak_list_.exists(cnt))
					{
						NakListEntry* ne =  NakListEntry::alloc();
						ne->init(this,cnt);
						nak_list_.addEntry(ne);
					}
			}	
		lvr = (lastSeqNo +1)% SEQ_NO_SIZE;
		 if( ipPacket ) 
		{
			hdr_rlp* hdrrlp1 = hdr_rlp::access(ipPacket);
			last = hdrrlp1->l_seq;
			if(!hdrrlp1->sent && !hdrrlp1->corrupted)reassembleBuffer.enque(ipPacket);
			readyToSend = true;
			ipPacket = 0;
		}
		}

		if(readyToSend)
		{
			int len_ = reassembleBuffer.length();
			for(int i =0; i < len_;++i)
			{
				Packet* pt = reassembleBuffer.deque();
				hdr_rlp* hdrrlp1 = hdr_rlp::access(pt);
				if(  !hdrrlp1->corrupted && !hdrrlp1->sent && compare(hdrrlp1->l_seq,lvn))
				{
					hdr_tcp* tcph = hdr_tcp::access(pt);
					//(print)out_<<tcph->seqno()<<"  "<<Scheduler::instance().clock()<<"  2"<<endl;	
					dst->recv(pt,(Handler*)0);
					hdrrlp1->sent = true;
				}
				else reassembleBuffer.enque(pt);
			}
		}
			nak_list_.process();
		Packet::free(pkt);
	return;
  }
   if(typeOfPacket == PT_RLP && !hdrcmn->error_)
   {
  	 hdr_rlp* hdr = hdr_rlp::access(pkt);
	currIpPacket = hdr->ippkt;
	 if((ipPacket  != currIpPacket) && currIpPacket  && compare(lastSeqNo, (hdr-> l_seq)%SEQ_NO_SIZE)) 
		{
			if(ipPacket) 
			{
			hdr_rlp* hdar = hdr_rlp::access(ipPacket);
			if(hdar->txmtd)
				{		
						hdr_rlp* hdrrlp1 = hdr_rlp::access(ipPacket);
						last = hdrrlp1->l_seq;
						if(!hdrrlp1->sent && !hdrrlp1->corrupted)
						{
							reassembleBuffer.enque(ipPacket);
							readyToSend = true;
							lastSeqNo = hdrrlp1->l_seq;
						}
							ipPacket = 0;
				}
			}
			ipPacket = currIpPacket;
		}
		if(  ipPacket && compare(lastSeqNo, (hdr-> l_seq+1)%SEQ_NO_SIZE) ) 
		 //if( (compare(lastSeqNo, (hdr-> l_seq)%SEQ_NO_SIZE)) && ipPacket ) 
		{
			hdr_rlp* hdar = hdr_rlp::access(ipPacket);
			if(hdar->txmtd)
			{
			if(!hdar->sent && !hdar->corrupted)
			{
				lastSeqNo = hdar->l_seq;
				reassembleBuffer.enque(ipPacket);
				readyToSend = true;
			}
			ipPacket = 0;
			}
		}
	   if(!(hdr->rexmit))
		{
			if(hdr->l_seq == lvr)
				{
					if(lvr ==lvn)
						{
							lvn = ((lvn+1)%(SEQ_NO_SIZE));
							lvr = ((lvr+1) % (SEQ_NO_SIZE)); 
							//out_<<hdr->l_seq<<"                "<<Scheduler::instance().clock()<<"           0"<<endl;	
						}
					if(lvr != lvn)  
					{
						lvr = ((lvr+1) % (SEQ_NO_SIZE)); 
						//out_<<hdr->l_seq<<"                "<<Scheduler::instance().clock()<<"           0"<<endl;
					}
				}
			if(compare(lvr,hdr->l_seq))
				{ 
					for (int cnt = lvr;compare(cnt,((hdr->l_seq))); cnt= ((cnt+1)%SEQ_NO_SIZE))
						{
							NakListEntry* ne =  NakListEntry::alloc();
							ne->init(this,cnt);
							nak_list_.addEntry(ne);	
						}
					lvr = (hdr->l_seq + 1)%SEQ_NO_SIZE;
					//out_<<hdr->l_seq<<"                "<<Scheduler::instance().clock()<<"           0"<<endl;
				}
			if(! hdr->last_seg) Packet::free(pkt); 
		}
	 if((hdr->rexmit) || ((!(hdr->rexmit)) && (compare(hdr->l_seq,lvn))))
		{
			if(hdr->rexmit && !compare(hdr->l_seq,lvn ) && hdr->ippkt )
			{
					bool check = false;
				int len_ = reassembleBuffer.length();
				for(int i =0; i < len_;++i)
					{
						Packet* pt = reassembleBuffer.deque();
						if(hdr->ippkt == pt)					
						{
							check = true;
						}
						reassembleBuffer.enque(pt);
					}
				if ( ipPacket == hdr->ippkt) check = true;
				if(!check) 
					{
						hdr_rlp* hdr_rlp3 = hdr_rlp::access(hdr->ippkt);
						if( !hdr_rlp3->sent && !hdr_rlp3->corrupted && hdr_rlp3->txmtd)	{reassembleBuffer.enque(hdr->ippkt);}
					}
			 len_ = reassembleBuffer.length();
			}
			if(!compare(hdr->l_seq,lvn ) && compare(hdr->l_seq,lvr))
				{
					if(hdr->l_seq == lvn) 
						{
							nak_list_.remove(hdr->l_seq);// out_<<hdr->l_seq<<"                "<<Scheduler::instance().clock()<<"           0"<<endl;;
							lvn = (hdr->l_seq+1)%SEQ_NO_SIZE;
							for(int cnt = ((hdr->l_seq+1)%SEQ_NO_SIZE) ;(compare(cnt,((lvr)%SEQ_NO_SIZE)) && !(nak_list_.exists(cnt)) ); cnt= ((cnt+1)%SEQ_NO_SIZE), lvn =cnt);
						}
					if(compare(lvn,hdr->l_seq))
						{
							nak_list_.remove(hdr->l_seq);//out_<<hdr->l_seq<<"                "<<Scheduler::instance().clock()<<"           0"<<endl; 
						}	 
					
				}
		 Packet::free(pkt); 
		}
   }
    
     if(typeOfPacket ==  PT_RLP_NAK && !hdrcmn->error_)
	{
		hdr_rlp* hdr_rlp1 = hdr_rlp::access(pkt);
		hdr_cmn* hdr_cmn1 = hdr_cmn::access(pkt);
		Packet* c = allocpkt(size_);	     	
		hdr_ip* hdrip2 = hdr_ip::access(c);
		hdr_rlp* hdr_rlp2 = hdr_rlp::access(c);
      		hdr_rlp2->l_seq=hdr_rlp1->l_seq;
		hdr_rlp2->rexmit = true;
		hdr_rlp2->ippkt = 0;
		hdrip2->flowid()= 102;
		for(int cnt = 0; cnt <30 && (ip[cnt]);++cnt)
		{
			if(compare(ipFirst[cnt], (hdr_rlp1->l_seq+1)%SEQ_NO_SIZE) && compare(hdr_rlp1->l_seq, (ipLast[cnt]+1)%SEQ_NO_SIZE) ) 
			{
				hdr_cmn* hdarcmn = hdr_cmn::access(ip[cnt]);
				if(packnum[cnt] == hdarcmn->uid())hdr_rlp2->ippkt = ip[cnt]; 
				continue;
			}
		} 
		rexmitSendBuffer.enque(c);   
		Packet::free(pkt);
	}
	else if ((typeOfPacket != PT_RLP) &&  (typeOfPacket != PT_RLP_NAK) && (typeOfPacket != PT_RLP_IDLE))
	{
	ipBuffer.enque(pkt);
	ipq += hdrcmn->size();
	}
	if( !hdrcmn->error_ && ((typeOfPacket == PT_RLP) ||  (typeOfPacket == PT_RLP_NAK) || (typeOfPacket == PT_RLP_IDLE))) nak_list_.process();
	if( !hdrcmn->error_ && ((typeOfPacket == PT_RLP) ||  (typeOfPacket == PT_RLP_NAK) || (typeOfPacket == PT_RLP_IDLE))) 
	{  
		int len_ = reassembleBuffer.length();
		for(int i =0; i < len_;++i)
			{
				Packet* pt = reassembleBuffer.deque();
				hdr_rlp* hdrrlp1 = hdr_rlp::access(pt);
				if(  !hdrrlp1->corrupted && !hdrrlp1->sent && compare(hdrrlp1->l_seq,lvn))
				{
					
						hdr_tcp* tcph = hdr_tcp::access(pt);
						//(print)out_<<tcph->seqno()<<"  "<<Scheduler::instance().clock()<<"  2"<<endl;	
						dst->recv(pt,(Handler*)0); 
						hdrrlp1->sent = true;
				}
			 if(!hdrrlp1->corrupted && !hdrrlp1->sent){ reassembleBuffer.enque(pt);}
			}
		if(reassembleBuffer.length() == 0) readyToSend = false; 
	}
}


void Rlp_Timer::expire(Event *e)
{
  a_->timeout(0);
} 


void RlpAgent::timeout(int)
{
	if(nakSendBuffer.length() != 0 ) send(nakSendBuffer.deque(),0);
	else if(rexmitSendBuffer.length() != 0) send(rexmitSendBuffer.deque(),0);
	else if((ipBuffer.length() != 0) || currSendIpPacket)
	{
		if(currSendIpPacket) 
		{
			Packet* paket = getPacket();
			hdr_rlp* hdr_rlp1 = hdr_rlp::access(paket);
			//out_<<hdr_rlp1->l_seq<<"                "<< Scheduler::instance().clock()<<"           2"<<endl;
			send(paket,0);
		}
		else if (!currSendIpPacket && (ipBuffer.length() != 0))
		{
			currSendIpPacket = ipBuffer.deque();
			begun = false;
			Packet* paket = getPacket();
			//hdr_rlp* hdr_rlp1 = hdr_rlp::access(paket);
			//out_<<hdr_rlp1->l_seq<<"                "<< Scheduler::instance().clock()<<"           2"<<endl;
			send(paket,0);
		}
	}
	else if(ifIdle)
	{
		Packet* c = allocpkt();
		hdr_cmn* hdrcomn = hdr_cmn::access(c);
		hdr_ip* hdrip = hdr_ip::access(c);
		hdrip->flowid() = 104;
		hdrcomn->ptype() = PT_RLP_IDLE;
		hdr_rlp* hdr_rlp1 = hdr_rlp::access(c);
     		hdr_rlp1->l_seq= -1; 
		//lvs = ((lvs+1)%(SEQ_NO_SIZE));
		send(c,0);
	}
	rlp_timer_.resched(0.02);
}

void RlpAgent::sendPkt(Packet* pt)
{
	nakSendBuffer.enque(pt);
}

void NakList::addEntry(NakListEntry* ne) 
{

	if (!tail_) head_= tail_= ne;
	else 
	{
		tail_->next_= ne;
		tail_= ne;
	}	
	tail_->next_= 0;
	++len_;
}

NakListEntry* NakList::getNext() 
{ 
  	if (!iter_) return 0;
	NakListEntry* tmp = iter_; iter_ = iter_->next_;
	return tmp;
}

void NakList::process()
{
	int nak_count[3] = {1, 2, 3};
	if(len_ == 0) return; 
	for(NakListEntry *p= head_; p;)
	{	
		if( (p->roundcount() > a_->NUM_ROUNDS) && (p->rt() >0))
		{
				p->decRT();
				if(p->rt() == 0)
				{
					NakListEntry* temp = p;//RlpAgent::out_<<p->seqno()<<"                "<<Scheduler::instance().clock()<<"              1"<<endl;
					p= p->next_;
					a_->removeCorrupted(temp->seqno());
					remove(temp->seqno()); 
					for(int cnt = (a_->getLVN()) ;(a_->compare(cnt,((a_->getlvr())%SEQ_NO_SIZE)) && !(exists(cnt)) ); cnt= ((cnt+1)%SEQ_NO_SIZE), a_->setLVN(cnt)) {}
				}
				else p = p->next_;
		}
		else
		{
			if ((p->roundcount() <= a_->NUM_ROUNDS) && (p->rt() > 0))
			{
				p->decRT();
				if((p->rt() == 0) || (p->roundcount() == 1) )
				{
					for(int cnt = 1; cnt <= nak_count[p->roundcount()-1]; ++cnt){p->sendNak();}
					p->setRT();
					p->incRC(); 
				}
				p = p->next_;
			}
		}
	}
}
bool NakList::remove(int sno)
{
	for (NakListEntry *p= head_, *pp = 0; p; pp= p, p= p->next_) 
	{	
		if (p->seqno() == sno )
		{
				if(!pp)
				{
					if (!head_) return false;
					head_= p->next_; 
		                        if (p == tail_) head_= tail_= 0;
		                       --len_;
					NakListEntry::free(p);
					return true;
				}
				else
				{
					if (p == tail_) tail_= pp;
					pp->next_= p->next_;
					--len_;
					NakListEntry::free(p);
					return true;
				}
		}
	}
	return false;
} 

void NakListEntry::sendNak()
{
  	Packet* c = a_->getNewPacket();
        hdr_cmn* hdrcomn = hdr_cmn::access(c);
	hdrcomn->ptype() = PT_RLP_NAK;
	hdr_ip* hdrip = hdr_ip::access(c);
	hdrip->flowid() = 101;
	hdr_rlp* hdr_rlp1 = hdr_rlp::access(c);
     	hdr_rlp1->l_seq=seq_no_; 
	a_->sendPkt(c);		
}

NakListEntry* NakList::search(int snum)
{
	NakListEntry* p= head_;
	NakListEntry* pp = 0;
	for (; p; pp= p, p= p->next_) 
	{	
		if (p->seqno() == snum)
		{
			return p;
		}
	}
	return 0;
}

bool NakList::exists(int sno)
{
	NakListEntry* p= head_;
	NakListEntry* pp = 0;
	bool flag = false;
	for (; p; pp= p, p= p->next_) 
	{	
		if (p->seqno() == sno)
		{
			flag= true;
		}
	}
	return flag;
}

void NakList::print()
{
	NakListEntry* p= head_;
	NakListEntry* pp = 0;
	for (; p; pp= p, p= p->next_) 
	{	
		cout<<p->seqno()<<"-";
	}
	if(head_ != 0)cout<<'\n';
}

int NakList::size()
{
	NakListEntry* p= head_;
	NakListEntry* pp = 0;
	int count = 0;
	for (; p; pp= p, p= p->next_) ++count;
	return count;
}

bool RlpAgent::compare(int a, int b)
{	
        if (a== b) return false;
	if( a < (SEQ_NO_SIZE/2))
	{
		 if ((b >  a ) && (b < (a+(SEQ_NO_SIZE/2)-1))) return true; 
		else return false;
	}
	 if( a >=  (SEQ_NO_SIZE/2) )
	{
		if ((b < a) && (b >= a- (SEQ_NO_SIZE/2))) return false;
		 else return true;
	}

}

void NakListEntry::setRT()
{
	int nak_count[3] = {1,2,3};
	rexmit_timer_ = rlp_delay + *(nak_count +round_counter_ - 1 );
}

inline void NakListEntry::free(NakListEntry* p)
{
			p->next_ = free_;
			free_ = p;
}

inline void NakListEntry::init(RlpAgent* a, int snum)
{
	a_ = a;
	seq_no_ = snum;
	round_counter_ = 1;
	rlp_delay = a_->RLP_DELAY;
	setRT();
}

inline NakListEntry* NakListEntry::alloc()
{	
	NakListEntry* p = free_;
	if (p != 0) 
	{
		free_ = p->next_;
	} 
	else 
	{
		p = new NakListEntry();
	}
	p->next_ = 0;
	return (p);
}

Packet* RlpAgent::getPacket()
{
	Packet* pkt = currSendIpPacket;
        hdr_cmn* hdrcmn = hdr_cmn::access(pkt);
   	hdr_ip* hdrip = hdr_ip::access(pkt);
        hdr_rlp* hdar = hdr_rlp::access(pkt);
        hdr_tcp* tcph = hdr_tcp::access(pkt);
	hdar->txmtd = false;
   	packet_t typeOfPacket = hdrcmn->ptype();
	double psize=hdrcmn->size(); 
    	packet_t org_ptype_=hdrcmn->ptype();
	hdar->packet_type_ = org_ptype_;
	int size_ = (int)((rlprate_ *20)/8);
	psize -= extra;
	if(!begun)
	{
		ip[ipCount] = pkt;
		ipFirst[ipCount] = lvs;
		packnum[ipCount] = hdrcmn->uid();
		if(hdrcmn->ptype() ==PT_TCP)
		{
		//(print)out_<<tcph->seqno()<<"  "<<Scheduler::instance().clock()<<"  0"<<endl;	
		}
		if(hdrcmn->ptype() ==PT_ACK)
		{
		//if(print)out_<<tcph->seqno()<<"  "<<Scheduler::instance().clock()<<"  1"<<endl;	
		}
		hdar->first = lvs;
		ipq -= extra;
	}
	Packet* c = 0;
	if(psize -done >size_ -5)
	{
		c = allocpkt();
      		hdr_rlp* hdr_rlp1 = hdr_rlp::access(c);
		hdr_cmn* hdrcmn1 = hdr_cmn::access(c);
		hdrcmn1->size() = size_;
      		hdr_rlp1->l_seq=lvs;
		if(!begun) hdr_rlp1->first  = lvs;
		hdr_rlp1->IP = *hdrip;
		hdr_rlp1->rexmit = false;
		hdr_rlp1->last_seg = false;
                hdr_rlp1->ippkt = pkt;
                ipLast[ipCount] = lvs;
      		hdr_ip* hdr_ip1 = hdr_ip::access(c);
      		int flow_ID_ = hdrip->flowid();
      		hdr_ip1->flowid() = 103;
		hdr_ip1->dst() = dst_;
	        hdr_rlp1->packet_type_=org_ptype_;
		lvs = ((lvs+1)%(SEQ_NO_SIZE));
		done += size_ -5;
		ipq -= (size_ -5);
	}
	else
	{
		c = allocpkt();
    		hdr_rlp* hdr2 = hdr_rlp::access(c);
		hdr_cmn* hdrcmn1 = hdr_cmn::access(c);
		hdrcmn1->size() = size_;
    		hdr2->last_seg = true;
    		hdr2->l_seq=lvs; 
                ipLast[ipCount] = lvs;
		if(ipCount <29) ++ ipCount;
		else ipCount = 0;
		lvs = ((lvs+1)%(SEQ_NO_SIZE));
    		hdr2->IP = *hdrip;
		hdr2->rexmit = false;
   	 	hdr2->pkt_size=(int)psize;
   		hdr2->ippkt = pkt;
		hdr2->isLast = true;
		hdr_ip* hdr_ip1 = hdr_ip::access(c);
     	 	int flow_ID_ = hdrip->flowid();
     	 	hdr_ip1->flowid() = 103;
		hdr_ip1->dst() = dst_;
      		hdr2->packet_type_=org_ptype_;
		hdrcmn->ptype() = type_;
		//hdrip->dst() = dst_;
		hdr_rlp* header = hdr_rlp::access(pkt);
		header->txmtd = true;
		header->last = ((lvs-1)%(SEQ_NO_SIZE));
		header->l_seq  = ((lvs-1)%(SEQ_NO_SIZE));
        	header->packet_type_ = org_ptype_;
		if(ipBuffer.length() != 0) extra = size_ -5 - psize + done ;
		else extra = 0;
		ipq -= (psize - done);
		done = 0;
		currSendIpPacket = 0;
	}
	if(!begun) begun = true;
	return c;
}


void RlpAgent::removeCorrupted(int seq)
{
	bool check = false;
	int len_ = reassembleBuffer.length();
	for(int i =0; i < len_;++i)
	{
		Packet* pt = reassembleBuffer.deque();
		hdr_rlp* hdr_rlp1 = hdr_rlp::access(pt);
		int las = hdr_rlp1->l_seq;
		int fir = hdr_rlp1->first;
		if((!compare(fir,(seq+1)%SEQ_NO_SIZE) || !compare(seq, (las+1)%SEQ_NO_SIZE)) && !hdr_rlp1->corrupted) {reassembleBuffer.enque(pt);}
		else 
		{
			hdr_rlp* hdrrlp3 = hdr_rlp::access(pt);
			hdrrlp3->corrupted = true;
			hdr_tcp* tcph = hdr_tcp::access(pt);
			//if(print)out_<<tcph->seqno()<<"                "<<Scheduler::instance().clock()<<"           3"<<endl;
			check = true;
		}
	}
	if(ipPacket)
	{
	hdr_rlp* hdr_rlp1 = hdr_rlp::access(ipPacket);
	int las = hdr_rlp1->l_seq;
	int fir = hdr_rlp1->first;
	if((compare(fir,(seq+1)%SEQ_NO_SIZE) && compare(seq, (las+1)%SEQ_NO_SIZE)) && !hdr_rlp1->corrupted)
		{
			hdr_rlp* hdrrlp3 = hdr_rlp::access(ipPacket);
			hdrrlp3->corrupted = true;
			hdr_tcp* tcph = hdr_tcp::access(ipPacket);
			//if(print) out_<<tcph->seqno()<<"                "<<Scheduler::instance().clock()<<"           3"<<endl;
			ipPacket = 0;
			check = true;
		}
	}
}
