 /*
 * File: rlp.h
 * Date: November 14, 2003
 * Author: Vikas Paliwal
 * Email: vpaliwal@sce.carleton.ca
 * Header file for the attributes and methods for RLP Agent
 */


#ifndef ns_rlp_h
#define ns_rlp_h

#include <fstream.h>
#include "agent.h"
#include "tclcl.h"
#include "packet.h"
#include "ip.h"
#include "tcp.h"
#include "rtp.h"
#include "queue.h"
#include "errmodel.h"

#define SEQ_NO_SIZE 4096

class RlpAgent;


class Rlp_Timer : public TimerHandler
{
	public:
		Rlp_Timer(RlpAgent *a) : TimerHandler() { a_ = a; }
		virtual void expire(Event *e);
	protected:
		RlpAgent *a_;
};

class NakListEntry
{
	public:
		NakListEntry(){}
		inline void init(RlpAgent*, int);
		int seqno(){return seq_no_;}
		NakListEntry* next_;
		inline int roundcount(){return round_counter_ ;}
		inline void incRC(){++round_counter_;}
		inline int rt(){return rexmit_timer_;}
		inline int decRT(){--rexmit_timer_;}
		void setRT();
		double sendTime;
		void sendNak();
		static inline NakListEntry* alloc();
		static inline void free(NakListEntry*);		int round_counter_;		int rexmit_timer_;
	protected:

		RlpAgent *a_;
		static NakListEntry* free_;
		int seq_no_;

		int rlp_delay;
		int* NAK_COUNT;
};

class NakList : public TclObject
{
	public:
		NakList() : head_(0), tail_(0), len_(0) {resetIterator(); }
		void init(RlpAgent* a){ a_ =a;}
		virtual int length() const { return (len_); }
		virtual void addEntry(NakListEntry*) ;
		virtual bool remove(int);
		virtual void process();
		virtual void print();
		int size();
		NakListEntry* search(int);
		virtual bool exists(int);
           	NakListEntry* head() { return head_; }
		NakListEntry* tail() { return tail_; }
	        void resetIterator() {iter_ = head_;}
        	NakListEntry* getNext() ;
	protected:
		RlpAgent *a_;
		NakListEntry* head_;
		NakListEntry* tail_;
		NakListEntry* iter_;
		int len_;
};

struct hdr_rlp
{
	  bool last_seg;
	  bool rexmit;
          Packet* ippkt;
	  int l_seq;
	  static int off_rlp_;
	  static int off_rtp_;
	  static int off_tcp_;
	  int pkt_size;
	  struct hdr_ip IP;
	  int last;
	  int first;
	  bool isLast;
	  bool txmtd;
	  bool sent;
	  bool corrupted;
	  packet_t packet_type_;
	  double sendTime;
	  int& seqno()
	  {
			return (l_seq);
	  }
  	  inline static hdr_rlp* access(Packet* p)
  	  {
		   return (hdr_rlp*) p->access(off_rlp_);
  	  }
};


class RlpAgent : public Agent
{
	 public:
		  int RLP_DELAY;
  		  static ofstream out_;
		  int NUM_ROUNDS;
		  RlpAgent();
		 // static const int nak_count[] = {1, 2, 3};
		  int command(int argc, const char*const* argv);
		  void recv(Packet*, Handler*);
		  void sendPkt(Packet*);
		  void timeout(int);
		  inline bool compare(int,int);
		  int getlvr(){return lvr;}
		  int getLVN(){return lvn;}
		  void setLVN(int lvnum){lvn = lvnum;}
		  void process();
		  void setNumRounds(int i){ NUM_ROUNDS =i;}
		  int getNumRounds(){return NUM_ROUNDS;}
		  void removeCorrupted(int);
		 Packet* getPacket();
		Packet* getNewPacket(){return allocpkt();}
	 protected:
		  int lvn;
		  int lvr;
		  int lvs;
                  double ipq;
		  int lastSeqNo;
		  int last;
		  double rlprate_;
		  ErrorModel* em;
		  int print;
		  bool ifIdle;
		  bool readyToSend;
		  Packet* ipPacket;
		  Packet* currIpPacket;
		  Packet* currSendIpPacket;
		  PacketQueue sendBuffer;
		  PacketQueue nakSendBuffer;
		  PacketQueue rexmitSendBuffer;
		  PacketQueue reassembleBuffer;
		  PacketQueue ipBuffer;
		  NakList nak_list_;
		  Rlp_Timer rlp_timer_;
		  NsObject* dst;
		  NakListEntry* ne;
	          int ipCounter;
		  int ipCounter1;
		 Packet* ip[30];
		  int packnum[30];
		  int ipFirst[30];
                 int ipLast[30];
		 int ipCount; 
		 double extra;
		 double done;
		 bool begun;
};

#endif
