#Create a simulator object
set ns [new Simulator]
$ns color 1 black
$ns color 101 red
$ns color 102 blue
$ns color 103 green
$ns color 104 white
set numDevices 4

#Open a trace file
set nf [open out.nam w]

# Trace Options
set tracefd  [open set.tr w]


#Define a 'finish' procedure
proc finish {} {
        global ns nf  tracefd
        $ns flush-trace
	close $tracefd
        close $nf
        exec nam out.nam &
        exit 0
}

#Create a set of matching nodes (actually 2set nodes - m($i),mr($i) is the higher/lower layer pair of the CDMA device
#and the fr($i)/f($i) is the lower/higher layer of the basestation/network
#the connection between mid1 and mid2 simluates "the radio channel"
for {set i 0} {$i < $numDevices} {incr i} {
  set m($i) [$ns node]
  set f($i) [$ns node]
  set mr($i) [$ns node]
  $mr($i) color "chocolate"
  $mr($i) shape "square"
  set fr($i) [$ns node]
  $fr($i) color "chocolate"
  $fr($i) shape "square"
  }
set mid1 [$ns node]
set mid2 [$ns node]
  
#Connect the nodes with two links
$ns trace-all $tracefd 
$ns namtrace-all $nf

for {set i 0} {$i < $numDevices} {incr i} {
  $ns duplex-link $m($i) $mr($i) 1Mb 10ms DropTail
  $ns duplex-link $f($i) $fr($i) 1Mb 10ms DropTail
  $ns duplex-link $mr($i) $mid1 1Mb 10ms DropTail
  $ns duplex-link $fr($i) $mid2 1Mb 10ms DropTail
  }  

#Ensure that the link speed here is greater than the
#link speed in the rlp model or else the speed here will
#be the limiting factor instead
$ns duplex-link $mid1 $mid2 200Mb 30ms DropTail
$ns duplex-link-op $mid1 $mid2 color "chocolate"
$ns duplex-link-op $mid1 $mid2 label "radio link"
$ns duplex-link-op $mid1 $mid2 orient right

# Create a loss model - error model rate_ variable describes the %loss the stream will
# experience. 1.00 = full loss and 0.05 = 5% loss.
set em_ [new ErrorModel]
$em_ set rate_ 0.0
#$ns link-lossmodel $em_ $n2 $n3
#these ones below are actually not used.  They're 2-state error models to
#produce bursty link errors
#		set j 0
#		set tmp1($j) [new ErrorModel/Uniform 0 pkt]
#		set tmp2($j) [new ErrorModel/Uniform 1 pkt]
#		set m_states($j) [ list $tmp1($j) $tmp2($j)]
#		set m_periods($j) [list 0.02 0.02]
#		set m_transmx($j) { {0.97 0.03}
#				{0.12 0.88}}
#		set m_trunit($j) pkt
#		set m_sttype($j) time
#		set m_nstates($j) 2
#		set m_nstart($j) [lindex $m_states($j) 0]
#		#set em_ [new ErrorModel/MultiState $m_states($j) $m_periods($j) $m_transmx($j) $m_trunit($j) $m_sttype($j) $m_nstates($j)  $m_nstart($j)]

$ns link-lossmodel $em_ $mid1 $mid2 
set em1_ [new ErrorModel]
$em1_ set rate_ 0.08
$ns link-lossmodel $em1_ $mid2 $mid1 


for {set i 0} {$i < $numDevices} {incr i} {
  set tcp_pass_m($i) [new Agent/Ppp]
  $ns attach-agent $mr($i) $tcp_pass_m($i)
  set tcp_rlp_m($i) [new Agent/Rlp]
  $ns attach-agent $mr($i) $tcp_rlp_m($i)
  $tcp_rlp_m($i) set rawrate_ 57.6
  $tcp_rlp_m($i) set rlp_delay 13

#note that each ftp transfer requires a pair of Agent/TCP and Agent/TCPSink
#Also, ftp model in ns-2 is a upload model instead of a download model
#if a download model is needed a "reverse" upload model from the other end
#should be used.
  set tcp_sender_m($i) [new Agent/TCP]
  ##$tcp_sender set packetSize_ 384
  ##$tcp_sender set class_ 1
  $ns attach-agent $m($i) $tcp_sender_m($i)
  $ns add-agent-trace $tcp_sender_m($i) tcp_agent
  set ftp_m($i) [new Application/FTP]
  $ftp_m($i) attach-agent $tcp_sender_m($i)
  $ftp_m($i) set type_ FTP
  set sink_m($i) [new Agent/TCPSink]
  $ns attach-agent $f($i) $sink_m($i)
  $ns add-agent-trace $sink_m($i) tcp_agent
  ##$ns connect $tcp $sink


  set tcp_rlp_f($i) [new Agent/Rlp]
  $ns attach-agent $fr($i) $tcp_rlp_f($i)
  set tcp_pass_f($i) [new Agent/Ppp]
  $ns attach-agent $fr($i) $tcp_pass_f($i)
  $tcp_rlp_f($i) set rawrate_ 130.6
  $tcp_rlp_f($i) set rlp_delay 13
  set tcp_sender_f($i) [new Agent/TCP]
  ##$tcp_sender set packetSize_ 384
  ##$tcp_sender set class_ 1
  $ns attach-agent $f($i) $tcp_sender_f($i)
#  $ns add-agent-trace $tcp_sender_f($i) tcp_agent
  set ftp_f($i) [new Application/FTP]
  $ftp_f($i) attach-agent $tcp_sender_f($i)
  $ftp_f($i) set type_ FTP
  set sink_f($i) [new Agent/TCPSink]
  $ns attach-agent $m($i) $sink_f($i)
#  $ns add-agent-trace $sink_f($i) tcp_agent
  ##$ns connect $tcp $sink
  }  



#these 6 lines connects the rlp and the ppp agents
#and lets the PPP agent know where the packets are
#destined after reassembly at the rlp layer.  In this case from mr0(mr1)
#the destination is f0(f1) and from fr0(fr1) the destination is m0(m1)
for {set i 0} {$i < $numDevices} {incr i} {
  $tcp_pass_m($i) dst $tcp_rlp_m($i)
  $tcp_rlp_m($i) dst $tcp_pass_m($i)
  $tcp_pass_m($i) destnode $f($i)
  $tcp_pass_f($i) dst $tcp_rlp_f($i)
  $tcp_rlp_f($i) dst $tcp_pass_f($i)
  $tcp_pass_f($i) destnode $m($i)
  }  

#this turns on the ppp function (so that the ppp just act as a pass
#through in cases where it's not enabled.
for {set i 0} {$i < $numDevices} {incr i} {
  [$mr($i) set classifier_] set isppp_ 1
  [$mr($i) set classifier_] ppphead $tcp_pass_m($i)
  [$fr($i) set classifier_] set isppp_ 1
  [$fr($i) set classifier_] ppphead $tcp_pass_f($i)

}

#The tcpsenders still needs to be connected to its sink
#same thing with the rlp connections.
for {set i 0} {$i < $numDevices} {incr i} {
  $ns connect $tcp_sender_m($i) $sink_m($i)
  $ns connect $tcp_sender_f($i) $sink_f($i)
  $ns connect $tcp_rlp_m($i) $tcp_rlp_f($i)
}


#Schedule events
#$ns at 0 "$ftp send 100000"
$ns at 0.0 "$ns trace-annotate \"Packet/Frame Color Legend:\""
$ns at 0.0 "$ns trace-annotate \"--------------------------\""
$ns at 0.0 "$ns trace-annotate \"BLACK : IP packet\""
$ns at 0.0 "$ns trace-annotate \"GREEN : RLP new data frame\""
$ns at 0.0 "$ns trace-annotate \"BLUE : RLP retransmitted data frame\""
$ns at 0.0 "$ns trace-annotate \"RED : RLP NAK frame\""
$ns at 0.0 "$ns trace-annotate \"WHITE : RLP idle frame\""
#$ns at 0.2 "$ftp_m(0) start"
#$ns at 0.9 "$ftp_f(0) start"
for {set i 0} {$i < $numDevices} {incr i} {
  $ns at 0.2 "$ftp_m($i) start"
  $ns at 0.4 "$ftp_f($i) start"
}
$ns at 20 "finish"
#$ns at 0.0 "$ns set-animation-rate 4ms"
#Run the simulation
$ns run


